/*++

Module Name:

    device.c - Device handling events for example driver.

Abstract:

   This file contains the device entry points and callbacks.

Environment:

    Kernel-mode Driver Framework

--*/

#include "driver.h"
#include "device.tmh"


#ifdef ALLOC_PRAGMA
#pragma alloc_text (PAGE, AIOWDM4CreateDevice)
#endif

const TBarData BlankBarData = { btNone, NULL, 0, {0, 0, }, };
const IOACTION BlankIOAction = { 'M', 'N', 0, 0, };

PDEVICE_OBJECT AIOWDMCreateCentralDevice(IN PDRIVER_OBJECT DriverObject)
{
    NTSTATUS Status;
    UNICODE_STRING NtDeviceName;
    UNICODE_STRING Win32DeviceName;
    PDEVICE_OBJECT pDevObj;
    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> %!FUNC! Entry ");
    RtlInitUnicodeString(&NtDeviceName, DEVICE_DEVICE_NAME);

    Status = IoCreateDevice(
        DriverObject,
        sizeof(DEVICE_CONTEXT),
        &NtDeviceName,
        FILE_DEVICE_UNKNOWN,
        0,
        FALSE, // Not exclusive
        &pDevObj
    );

    if (!NT_SUCCESS(Status))
    {
        TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> IoCreateDevice failed status %d\n", Status);
        return NULL; // Give up if create failed.
    }

    RtlInitUnicodeString(&Win32DeviceName, AIOWDM_DEVICE_NAME);
    Status = IoCreateSymbolicLink(&Win32DeviceName, &NtDeviceName);
    if (!NT_SUCCESS(Status)) // If we we couldn't create the link then abort installation.
    {
        IoDeleteDevice(pDevObj);
        TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> IoCreateSymbolicLink failed status %d\n", Status);
        return NULL;
    }

    PDEVICE_CONTEXT pdx = (PDEVICE_CONTEXT)pDevObj->DeviceExtension;
    pdx->StatusBar = BlankBarData;
    pdx->MainBar = BlankBarData;
    pdx->AuxBar = BlankBarData;
    pdx->bRealDevice = FALSE;

    return pDevObj;
}




NTSTATUS
AIOWDM4CreateDevice(_In_ PWDFDEVICE_INIT DeviceInit)
/*++
Routine Description:
    Worker routine called to create a device and its software resources.
Arguments:
    DeviceInit - Pointer to an opaque init structure. Memory for this
                    structure will be freed by the framework when the WdfDeviceCreate
                    succeeds. So don't access the structure after that point.
Return Value:
    NTSTATUS
--*/
{
    WDF_OBJECT_ATTRIBUTES deviceAttributes;
    PDEVICE_CONTEXT deviceContext;
    WDFDEVICE Device;
    NTSTATUS status;

    PAGED_CODE();
    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> %!FUNC! Enter");

    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&deviceAttributes, DEVICE_CONTEXT);
    WdfDeviceInitSetIoType(DeviceInit, WdfDeviceIoBuffered);

    status = WdfDeviceCreate(&DeviceInit, &deviceAttributes, &Device);
    if (!NT_SUCCESS(status)) {
        TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "WdfDeviceCreate failed with status %!STATUS!\n", status);
        goto funcexit;
    }
    //
  // Initialize the I/O Package and any Queues
    status = AIOWDM4QueueInitialize(Device);
    if (!NT_SUCCESS(status)) {
        TraceEvents(TRACE_LEVEL_ERROR, DBG_INIT, "AIOWDM4QueueInitialize failed %!STATUS!\n", status);
    }

    deviceContext = AIOWDM4DeviceGetContext(Device);

    //
    // Initialize the context.
    // stuff the device-specific fields

    deviceContext->WdfDeviceObject = Device;              
    //deviceContext->LowerDeviceObject = ; 

    PDEVICE_OBJECT deviceObject;
    deviceObject = WdfDeviceWdmGetPhysicalDevice(Device);
    deviceContext->Pdo = deviceObject;

    //deviceContext->RemoveLock = ;
    //deviceContext->ifname = ;
    //deviceContext->state = ;
    //deviceContext->prevstate = ;
    //deviceContext->devpower = ;
    //deviceContext->syspower = ;
    //deviceContext->PerfBoundary = ;
    //deviceContext->oldseq = ;
    //deviceContext->devcaps = ;
    //deviceContext->handles = ;
    //deviceContext->InterruptObject = ;

    deviceContext->bRealDevice = TRUE;
    deviceContext->StatusBar = BlankBarData;    
    deviceContext->MainBar = BlankBarData;      
    deviceContext->AuxBar = BlankBarData;       

    funcexit:
    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "<-- %!FUNC! Exit Status: %!STATUS!", status);

    return status;
}
