/*++

Module Name:

    device.h

Abstract:

    This file contains the device definitions.

Environment:

    Kernel-mode Driver Framework

--*/

#include "public.h"

EXTERN_C_START


typedef enum
{
	STOPPED,								// device stopped
	WORKING,								// started and working
	PENDINGSTOP,							// stop pending
	PENDINGREMOVE,							// remove pending
	SURPRISEREMOVED,						// removed by surprise
	REMOVED,								// removed
} DEVSTATE;

///////////////////////////////////////////////////////////////////////////////
// Device extension structure

typedef struct _IOACTION
{
    char BarID;
    char Operation;
    UCHAR Offset;
    UCHAR Data;
} IOACTION, * PIOACTION;


typedef enum _TBarType { btNone, btMemory, btIO, btMappedIO } TBarType;
typedef struct _TBarData
{
    TBarType BarType;
    PUCHAR Base;
    ULONG Size;
    PHYSICAL_ADDRESS UnmappedBase;
} TBarData, * TPBarData;


typedef struct _TCOSScheme
{
    UCHAR StartOff, Size, Stride;
} TCOSScheme, * TPCOSScheme;

typedef struct _TCARDINFO
{
    ULONG DeviceID;
    ULONG BaseAddress;
} TCARDINFO, * PCARDINFO;

typedef union
{
    UCHAR Data8;
    USHORT Data16;
    ULONG Data32;
}DATA;

typedef struct _TMEMOP
{
    ULONG Address;
    USHORT Padding0;
    UCHAR Flags;
    UCHAR Size;
    DATA Data;
} TMEMOP, * PMEMOP;

#define POF_RELATIVE 0x01
//
// The "device context" performs the same job as a "WDM device extension" in the driver frameworks
//
typedef struct _DEVICE_CONTEXT
{
	WDFDEVICE WdfDeviceObject;			// device object this extension belongs to
	WDFDEVICE LowerDeviceObject;		// next lower driver in same stack
	PDEVICE_OBJECT Pdo;						// the PDO
	IO_REMOVE_LOCK RemoveLock;				// removal control locking structure
	UNICODE_STRING ifname;					// interface name
	DEVSTATE state;							// current state of device
	DEVSTATE prevstate;						// state prior to removal query
	DEVICE_POWER_STATE devpower;			// current device power state
	SYSTEM_POWER_STATE syspower;			// current system power state
	DEVICE_POWER_STATE PerfBoundary;		// context restore very expensive from >= this power state
	POWER_SEQUENCE oldseq;					// old power sequence numbers
	DEVICE_CAPABILITIES devcaps;			// copy of most recent device capabilities
	LONG handles;							// # open handles
	WDFINTERRUPT InterruptObject;			// address of interrupt object

	BOOLEAN bRealDevice; //false if central device(just for global commands)
	ULONG DeviceID;							// The card's PCI(e) device ID or ISA virtual device ID
	TBarData StatusBar, MainBar, AuxBar;
	//IOACTION IAm;							// how to find out if the device is generating an IRQ
	//IOACTION Disable;						// how to disable IRQs
	//IOACTION Clear;						// how to clear IRQs
	//IOACTION Enable;						// how to enable IRQs
	ULONG InterruptOccurred;				// --J2H ToDo: change to per-handle flag for multi-process waitForIRQ
	ULONG InterruptRequested;
	WDFREQUEST PendingIrp;
	BOOLEAN HasPendingIrp;
	//long IRQIndex;
	ULONG COSSize;
	//ULONG Bench;
	TCOSScheme COSScheme;
	//BOOLEAN GotIRQ; //true if device has IRQ
	WDFQUEUE IRQQueue;

	WDFSPINLOCK RcvLock;
	WDFSPINLOCK Lock;
	WDFSPINLOCK SendLock;
	BUS_INTERFACE_STANDARD  BusInterface;
} DEVICE_CONTEXT, *PDEVICE_CONTEXT;

//
// This macro will generate an inline function called DeviceGetContext
// which will be used to get a pointer to the device context memory
// in a type safe manner.
//
WDF_DECLARE_CONTEXT_TYPE_WITH_NAME(DEVICE_CONTEXT, AIOWDM4DeviceGetContext)

//
// Function to initialize the device and its callbacks
//
NTSTATUS
AIOWDM4CreateDevice(
	_In_ PWDFDEVICE_INIT DeviceInit
    );

NTSTATUS
AIOWDM4GetDeviceInformation(
	IN OUT PDEVICE_CONTEXT deviceContext
);
EXTERN_C_END
