/*++

Module Name:

    driver.h

Abstract:

    This file contains the driver definitions.

Environment:

    Kernel-mode Driver Framework

--*/

#include <ntddk.h>
#include <wdf.h>
#include <initguid.h>

#include "device.h"
#include "queue.h"
#include "trace.h"

EXTERN_C_START

#ifdef _WIN64
typedef ULONGLONG UNATIVE;
#else
typedef ULONG UNATIVE;
#endif

#define maskDisableIRQ 0x00
#define maskEnableIRQ 0x09
#define IRQEnableOffset 0x69
#define IRQStatusOffset 0x0F
#define COSIRQLatchOffset 0x08

#define DEVICE_DEVICE_NAME L"\\Device\\AIOWDMCoreDevice"
#define AIOWDM_DEVICE_NAME L"\\DosDevices\\AIOWDMCore"
//
// WDFDRIVER Events
//

DRIVER_INITIALIZE DriverEntry;
EVT_WDF_DRIVER_DEVICE_ADD AIOWDM4EvtDeviceAdd;
EVT_WDF_OBJECT_CONTEXT_CLEANUP AIOWDM4EvtDriverContextCleanup;

EVT_WDF_INTERRUPT_ISR AIOWDM4EvtInterruptIsr;
EVT_WDF_INTERRUPT_DPC AIOWDM4EvtInterruptDpc;
EVT_WDF_INTERRUPT_ENABLE AIOWDM4EvtInterruptEnable;
EVT_WDF_INTERRUPT_DISABLE AIOWDM4EvtInterruptDisable;
NTSTATUS AIOWDM4CreateIRQQueue(_In_ WDFDEVICE Device);

NTSTATUS AIOWDM4EvtDeviceAdd(WDFDRIVER Driver, PWDFDEVICE_INIT DeviceInit);
NTSTATUS AIOWDM4EvtPrepareHardware(
    WDFDEVICE Device,
    WDFCMRESLIST ResourcesRaw,
    WDFCMRESLIST ResourcesTranslated
);
NTSTATUS
AIOWDM4EvtReleaseHardware(
    WDFDEVICE Device,
    WDFCMRESLIST ResourcesTranslated
);

NTSTATUS
AIOWDM4UnmapHWResources(
    _In_ OUT PDEVICE_CONTEXT deviceContext
);

NTSTATUS
AIOWDM4MapHWResources(
    _In_ OUT PDEVICE_CONTEXT deviceContext,
    _In_ WDFCMRESLIST  ResourcesRaw,
    _In_ WDFCMRESLIST  ResourcesTranslated
);

TBarData BarFromResource(CM_PARTIAL_RESOURCE_DESCRIPTOR& Resource);
void DelBar(TBarData& Bar);

EXTERN_C_END
