/*++

Module Name:

    queue.c

Abstract:

    This file contains the queue entry points and callbacks.

Environment:

    Kernel-mode Driver Framework

--*/

#include "driver.h"
#include "queue.tmh"


#ifdef ALLOC_PRAGMA
#pragma alloc_text (PAGE, AIOWDM4QueueInitialize)
#pragma alloc_text (PAGE, ForgetIRP)
#pragma alloc_text (PAGE, BarFromResource)
#pragma alloc_text (PAGE, DelBar)
#endif

#ifndef CTL_CODE
#pragma message("CTL_CODE undefined. Include winioctl.h or wdm.h")
#endif

#define FILE_DEVICE_KRNLDRVR 0x80ff

#define IOCTL_QUERY_INFO     CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x800, METHOD_BUFFERED, FILE_READ_ACCESS)
#define IOCTL_QUERY_BAR_BASE CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x801, METHOD_BUFFERED, FILE_READ_ACCESS)
#define IOCTL_IRQ_WAIT_BEGIN CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x810, METHOD_BUFFERED, FILE_READ_ACCESS)
#define IOCTL_IRQ_WAIT_ABORT CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x812, METHOD_BUFFERED, FILE_ANY_ACCESS )
#define IOCTL_IRQ_POLL       CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x814, METHOD_BUFFERED, FILE_READ_ACCESS)
#define IOCTL_SET_COS_SCHEME CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x816, METHOD_BUFFERED, FILE_READ_ACCESS)
#define IOCTL_FORGET_IRP     CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x880, METHOD_BUFFERED, FILE_WRITE_ACCESS)

#define IOCTL_PORT_IN        CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x820, METHOD_BUFFERED, FILE_READ_ACCESS )
#define IOCTL_PORT_INSTR     CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x821, METHOD_BUFFERED, FILE_READ_ACCESS )
#define IOCTL_PORT_OUT       CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x822, METHOD_BUFFERED, FILE_WRITE_ACCESS)
//#define IOCTL_PORT_OUTSTR    CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x823, METHOD_BUFFERED, FILE_WRITE_ACCESS)
#define IOCTL_MEM_IN         CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x824, METHOD_BUFFERED, FILE_READ_ACCESS )
#define IOCTL_MEM_OUT        CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x826, METHOD_BUFFERED, FILE_WRITE_ACCESS)
#define IOCTL_PORTMB_OUT     CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x82A, METHOD_BUFFERED, FILE_WRITE_ACCESS)

#define IOCTL_GET_BENCH      CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x830, METHOD_BUFFERED, FILE_READ_ACCESS )





#pragma PAGEDCODE

void ForgetIRP(PIRP IRP)
{
    PAGED_CODE();
    UNREFERENCED_PARAMETER(IRP);
    //This function is evil. However, if we complete the IRP, it'll ret into a thread that no
    //longer exists. This way, it'll remain pending indefinitely, consuming resources
    //including nonpaged memory... but the system won't bugcheck.

    //It would be really nice of us to delete the IRP without completing it, but we don't
    //know how, and furthermore don't believe there's a way(since it wouldn't be necessary
    //without this "ret into a thread that no longer exists" bug).
}

///*********************************************************************************
//*      CANCEL IRP
//**********************************************************************************/
//
//#pragma LOCKEDCODE
//
//VOID
//IRQGenCancelIrp(
//    IN PDEVICE_OBJECT DeviceObject,
//    IN PIRP Irp
//)
//{
//    PDEVICE_CONTEXT deviceContext = AIOWDM4DeviceGetContext(DeviceObject);
//
//    Irp->IoStatus.Information = 0;
//
//    if (Irp == deviceContext->PendingIrp)
//    {
//        deviceContext->HasPendingIrp = FALSE;
//        Irp->IoStatus.Information = 1;
//    }
//    IoReleaseCancelSpinLock(Irp->CancelIrql);
//
//    Irp->IoStatus.Status = STATUS_CANCELLED;
//    IoCompleteRequest(Irp, IO_NO_INCREMENT);
//}

///////////////////////////////////////////////////////////////////////////////
//BAR management

#pragma PAGEDCODE

TBarData BarFromResource(CM_PARTIAL_RESOURCE_DESCRIPTOR & Resource)
{
	PAGED_CODE();

	TBarData Result;
	RtlZeroMemory(&Result, sizeof(Result));

	switch ( Resource.Type )
	{
		case CmResourceTypePort:
			Result.Size = Resource.u.Port.Length;
			if ( Resource.Flags & CM_RESOURCE_PORT_IO )
			{
				Result.BarType = btIO;
				Result.UnmappedBase = Resource.u.Port.Start;
			}
			else
			{
				Result.BarType = btMappedIO;
				Result.UnmappedBase = Resource.u.Port.Start;
			}
		break;
		case CmResourceTypeMemory:
			Result.BarType = btMemory;
			Result.Size = Resource.u.Memory.Length;
			Result.UnmappedBase = Resource.u.Memory.Start;
		break;
		default:
			return Result;
		//break;
	}
	if ( Result.BarType == btIO )
		Result.Base = PUCHAR(Result.UnmappedBase.QuadPart);
	else
		Result.Base = PUCHAR(MmMapIoSpace(Result.UnmappedBase, Result.Size, MmNonCached))
	;

	return Result;
}

#pragma PAGEDCODE

void DelBar(TBarData & Bar)
{
	PAGED_CODE();

	if ( Bar.Base ) switch ( Bar.BarType )
	{
		case btMemory: case btMappedIO: MmUnmapIoSpace(Bar.Base, Bar.Size);
	}
	RtlZeroMemory(&Bar, sizeof(Bar));
}

///////////////////////////////////////////////////////////////////////////////
//BAR use

#pragma LOCKEDCODE

UCHAR InBar8(const TBarData Bar, ULONG Offset)
{
    UCHAR Result;
    PUCHAR Addr = (PUCHAR)(Bar.Base + Offset);
    if (Bar.BarType == btMemory)
        Result = READ_REGISTER_UCHAR(Addr);
    else
        Result = READ_PORT_UCHAR(Addr)
        ;
    return Result;
}

#pragma LOCKEDCODE

USHORT InBar16(const TBarData Bar, ULONG Offset)
{
    USHORT Result;
    PUSHORT Addr = (PUSHORT)(Bar.Base + Offset);
    if (Bar.BarType == btMemory)
        Result = READ_REGISTER_USHORT(Addr);
    else
        Result = READ_PORT_USHORT(Addr)
        ;
    return Result;
}

#pragma LOCKEDCODE

ULONG InBar32(const TBarData Bar, ULONG Offset)
{
    ULONG Result;
    PULONG Addr = (PULONG)(Bar.Base + Offset);
    if (Bar.BarType == btMemory)
        Result = READ_REGISTER_ULONG(Addr);
    else
        Result = READ_PORT_ULONG(Addr)
        ;
    return Result;
}


#pragma LOCKEDCODE

void OutBar8(const TBarData Bar, ULONG Offset, UCHAR Data)
{
    PUCHAR Addr = (PUCHAR)(Bar.Base + Offset);
    if (Bar.BarType == btMemory)
        WRITE_REGISTER_UCHAR(Addr, Data);
    else
        WRITE_PORT_UCHAR(Addr, Data)
        ;
}

#pragma LOCKEDCODE

void OutBar16(const TBarData Bar, ULONG Offset, USHORT Data)
{
    PUSHORT Addr = (PUSHORT)(Bar.Base + Offset);
    if (Bar.BarType == btMemory)
        WRITE_REGISTER_USHORT(Addr, Data);
    else
        WRITE_PORT_USHORT(Addr, Data)
        ;
}

#pragma LOCKEDCODE

void OutBar32(const TBarData Bar, ULONG Offset, ULONG Data)
{
    PULONG Addr = (PULONG)(Bar.Base + Offset);
    if (Bar.BarType == btMemory)
        WRITE_REGISTER_ULONG(Addr, Data);
    else
        WRITE_PORT_ULONG(Addr, Data)
        ;
}


#pragma LOCKEDCODE

VOID InCOSData(PDEVICE_CONTEXT pdx, PUCHAR pDst, size_t COSSize)
{
    ULONG PPISize = pdx->COSScheme.Size;
    ULONG Offset = pdx->COSScheme.StartOff;
    ULONG dStride = pdx->COSScheme.Stride - pdx->COSScheme.Size;
    while (COSSize--)
    {
        if (PPISize == 0)
        {
            PPISize = pdx->COSScheme.Size;
            Offset += dStride;
        }
        --PPISize;

        *(pDst++) = InBar8(pdx->MainBar, Offset++);
    }
}

NTSTATUS
AIOWDM4QueueInitialize(
    _In_ WDFDEVICE Device
)
/*++

Routine Description:

     The I/O dispatch callbacks for the frameworks device object
     are configured in this function.

     A single default I/O Queue is configured for parallel request
     processing, and a driver context memory allocation is created
     to hold our structure QUEUE_CONTEXT.

Arguments:

    Device - Handle to a framework device object.

Return Value:

    VOID

--*/
{
    WDFQUEUE queue;
    NTSTATUS status;
    WDF_IO_QUEUE_CONFIG queueConfig;

    PAGED_CODE();

    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> %!FUNC! Enter");

    //
    // Configure a default queue so that requests that are not
    // configure-fowarded using WdfDeviceConfigureRequestDispatching to goto
    // other queues get dispatched here.
    //
    WDF_IO_QUEUE_CONFIG_INIT_DEFAULT_QUEUE(
        &queueConfig,
        WdfIoQueueDispatchParallel
    );

    queueConfig.EvtIoDeviceControl = AIOWDM4EvtIoDeviceControl;
    queueConfig.EvtIoStop = AIOWDM4EvtIoStop;

    status = WdfIoQueueCreate(
        Device,
        &queueConfig,
        WDF_NO_OBJECT_ATTRIBUTES,
        &queue
    );

    if (!NT_SUCCESS(status)) {
        TraceEvents(TRACE_LEVEL_ERROR, TRACE_QUEUE, "WdfIoQueueCreate failed %!STATUS!", status);
        return status;
    }

    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "<-- %!FUNC! Exit");

    return status;
}


VOID
AIOWDM4EvtIoDeviceControl(
    _In_ WDFQUEUE Queue,
    _In_ WDFREQUEST Request,
    _In_ size_t OutputBufferLength,
    _In_ size_t InputBufferLength,
    _In_ ULONG IoControlCode
)
/*++

Routine Description:

    This event is invoked when the framework receives IRP_MJ_DEVICE_CONTROL request.

Arguments:

    Queue -  Handle to the framework queue object that is associated with the
             I/O request.

    Request - Handle to a framework request object.

    OutputBufferLength - Size of the output buffer in bytes

    InputBufferLength - Size of the input buffer in bytes

    IoControlCode - I/O control code.

Return Value:

    VOID

--*/
{
    NTSTATUS status = STATUS_SUCCESS;
    PVOID                   DataBuffer;
    size_t                  BufferLength;
    ULONG                   infolen = 0;

    TraceEvents(TRACE_LEVEL_VERBOSE, TRACE_QUEUE, "--> %!FUNC! Enter");

    switch (IoControlCode) {
    case IOCTL_QUERY_INFO:
    {
        TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "IOCTL_QUERY_INFO");

        // Since the IOCTL is buffered, WdfRequestRetrieveOutputBuffer &
        // WdfRequestRetrieveInputBuffer return the same buffer pointer.
        // So make sure you read all the information you need from
        // the buffer before you write to it.

        PDEVICE_CONTEXT deviceContext = AIOWDM4DeviceGetContext(WdfIoQueueGetDevice(Queue));

        status = WdfRequestRetrieveOutputBuffer(Request, sizeof(TCARDINFO), &DataBuffer, &BufferLength);
        if (!NT_SUCCESS(status)) {
            TraceEvents(TRACE_LEVEL_ERROR, TRACE_QUEUE, "WdfRequestRetrieveOutputBuffer failed 0x%x\n", status);
            WdfRequestComplete(Request, status);
            return;
        }

        if (!deviceContext->bRealDevice)
        {
            status = STATUS_OBJECT_TYPE_MISMATCH;
            break;
        }

        if (OutputBufferLength < sizeof(TCARDINFO))
        {
            status = STATUS_INVALID_PARAMETER;
            break;
        }

        PCARDINFO CIBuf = (PCARDINFO)DataBuffer;
        CIBuf->DeviceID = deviceContext->DeviceID;
        CIBuf->BaseAddress = (ULONG)(UNATIVE)(deviceContext->MainBar.Base);
        TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "DeviceID: %04X BAR[2]=%04x", CIBuf->DeviceID, CIBuf->BaseAddress);

        infolen = sizeof(*CIBuf);

        status = STATUS_SUCCESS;

        break;
    }						// IOCTL_QUERY_INFO

    case IOCTL_QUERY_BAR_BASE:				// IOCTL_QUERY_BAR_BASE
    {
        TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "IOCTL_QUERY_BAR_BASE");

        status = WdfRequestRetrieveOutputBuffer(Request, sizeof(TCARDINFO), &DataBuffer, &BufferLength);
        if (!NT_SUCCESS(status)) {
            TraceEvents(TRACE_LEVEL_ERROR, TRACE_QUEUE, "WdfRequestRetrieveOutputBuffer failed 0x%x\n", status);
            WdfRequestComplete(Request, status);
            return;
        }
        PDEVICE_CONTEXT deviceContext = AIOWDM4DeviceGetContext(WdfIoQueueGetDevice(Queue));

        if (!deviceContext->bRealDevice)
        {
            status = STATUS_OBJECT_TYPE_MISMATCH;
            break;
        }

        if (OutputBufferLength < sizeof(ULONG))
        {
            status = STATUS_INVALID_PARAMETER;
            break;
        }

        PULONG pBuf = (PULONG)DataBuffer;
        switch (*pBuf)
        {
        case 1:
            *pBuf = (ULONG)(UNATIVE)(deviceContext->StatusBar.Base);
            break;
        case 2:
            *pBuf = (ULONG)(UNATIVE)(deviceContext->MainBar.Base);
            break;
        case 3:
            *pBuf = (ULONG)(UNATIVE)(deviceContext->AuxBar.Base);
            break;
        default:
            *pBuf = 0;
            break;
        }
        infolen = sizeof(*pBuf);
        status = STATUS_SUCCESS;

        break;
    }						// IOCTL_QUERY_BAR_BASE

    case IOCTL_PORT_IN:				// code == 0x820
    {
        TraceEvents(TRACE_LEVEL_VERBOSE, TRACE_QUEUE, "IOCTL_PORT_IN");

        if ((InputBufferLength < sizeof(TPORTOP)) || OutputBufferLength < sizeof(TPORTOP))
        {
            status = STATUS_INVALID_PARAMETER;
            break;
        }

        // Since the IOCTL is buffered, WdfRequestRetrieveOutputBuffer &
        // WdfRequestRetrieveInputBuffer return the same buffer pointer.
        // So make sure you read all the information you need from
        // the buffer before you write to it.

        status = WdfRequestRetrieveOutputBuffer(Request, sizeof(TCARDINFO), &DataBuffer, &BufferLength);
        if (!NT_SUCCESS(status)) {
            TraceEvents(TRACE_LEVEL_ERROR, TRACE_QUEUE, "WdfRequestRetrieveOutputBuffer failed 0x%x\n", status);
            WdfRequestComplete(Request, status);
            return;
        }

        PDEVICE_CONTEXT deviceContext = AIOWDM4DeviceGetContext(WdfIoQueueGetDevice(Queue));
        PPORTOP POBuf = (PPORTOP)DataBuffer;
        if (!deviceContext->bRealDevice)
        {
            if (POBuf->Address != 0x61) status = STATUS_OBJECT_TYPE_MISMATCH;
            else TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "InPort of 0x61");
        }
        else {
            TBarData TarBar = deviceContext->MainBar;
            switch (POBuf->Size)
            {
            case 1:
                POBuf->Data.Data8 = InBar8(TarBar, POBuf->Address);
                TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "RelInPortB(%04x)=>%02hx", (ULONG)(UNATIVE)(TarBar.Base + POBuf->Address), POBuf->Data.Data8);
                break;
            case 2:
                POBuf->Data.Data16 = InBar16(TarBar, POBuf->Address);
                TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "RelInPort(%04x)=>%04x", (ULONG)(UNATIVE)(TarBar.Base + POBuf->Address), POBuf->Data.Data16);
                break;
            case 4:
                POBuf->Data.Data32 = InBar32(TarBar, POBuf->Address);
                TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "RelInPortL(%04x)=>%08lx", (ULONG)(UNATIVE)(TarBar.Base + POBuf->Address), POBuf->Data.Data32);
                break;
            }
        }
        infolen = sizeof(*POBuf);

        status = STATUS_SUCCESS;

        break;
    }						// IOCTL_PORT_IN

    case IOCTL_PORT_OUT:	// code == 0x822
    {
        TraceEvents(TRACE_LEVEL_VERBOSE, TRACE_QUEUE, "IOCTL_PORT_OUT");



        if (OutputBufferLength < sizeof(TPORTOP))
        {
            status = STATUS_INVALID_PARAMETER;
            break;
        }
        PDEVICE_CONTEXT deviceContext = AIOWDM4DeviceGetContext(WdfIoQueueGetDevice(Queue));
        if (!deviceContext->bRealDevice)
        {
            status = STATUS_OBJECT_TYPE_MISMATCH;
            break;
        }
        // Since the IOCTL is buffered, WdfRequestRetrieveOutputBuffer &
        // WdfRequestRetrieveInputBuffer return the same buffer pointer.
        // So make sure you read all the information you need from
        // the buffer before you write to it.

        status = WdfRequestRetrieveOutputBuffer(Request, sizeof(TCARDINFO), &DataBuffer, &BufferLength);
        if (!NT_SUCCESS(status)) {
            TraceEvents(TRACE_LEVEL_ERROR, TRACE_QUEUE, "WdfRequestRetrieveOutputBuffer failed 0x%x\n", status);
            WdfRequestComplete(Request, status);
            return;
        }

        PPORTOP POBuf = (PPORTOP)DataBuffer;
        TBarData TarBar = deviceContext->MainBar;
        switch (POBuf->Size)
        {
        case 1:
            OutBar8(TarBar, POBuf->Address, POBuf->Data.Data8);
            TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "RelOutPortB(%04x, %02hx)", (ULONG)(UNATIVE)(TarBar.Base + POBuf->Address), POBuf->Data.Data8);
            break;
        case 2:
            OutBar16(TarBar, POBuf->Address, POBuf->Data.Data16);
            TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "RelOutPort(%04x, %04x)", (ULONG)(UNATIVE)(TarBar.Base + POBuf->Address), POBuf->Data.Data16);
            break;
        case 4:
            OutBar32(TarBar, POBuf->Address, POBuf->Data.Data32);
            TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "RelOutPortL(%04x, %08lx)", (ULONG)(UNATIVE)(TarBar.Base + POBuf->Address), POBuf->Data.Data32);
            break;
        }
        status = STATUS_SUCCESS;
        infolen = sizeof(*POBuf);
        break;
    }						// IOCTL_PORT_OUT

    case IOCTL_IRQ_WAIT_BEGIN:				// IOCTL_IRQ_WAIT_BEGIN
    {
        TraceEvents(TRACE_LEVEL_VERBOSE, TRACE_QUEUE, "IOCTL_IRQ_WAIT_BEGIN");
        PDEVICE_CONTEXT deviceContext = AIOWDM4DeviceGetContext(WdfIoQueueGetDevice(Queue));
        if (!deviceContext->bRealDevice)
        {
            status = STATUS_OBJECT_TYPE_MISMATCH;
            break;
        }

        if (deviceContext->HasPendingIrp == TRUE)
        {
            status = STATUS_TOO_MANY_COMMANDS;
        }
        else
            if (deviceContext->InterruptOccurred) // immediately complete request if an IRQ occurred while no one was pending for it
            {
                TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "Pending IRQ; IOCTL_IRQ_WAIT_BEGIN completed");

                deviceContext->InterruptOccurred = FALSE; // --J2H: Todo: change to per-handle list of this flag
                if (OutputBufferLength)
                {
                    status = WdfRequestRetrieveOutputBuffer(Request, sizeof(TCARDINFO), &DataBuffer, &BufferLength);
                    if (!NT_SUCCESS(status)) {
                        TraceEvents(TRACE_LEVEL_ERROR, TRACE_QUEUE, "WdfRequestRetrieveOutputBuffer failed 0x%x\n", status);
                        break;
                    }
                    InCOSData(deviceContext, PUCHAR(DataBuffer), OutputBufferLength);
                    infolen = (ULONG)OutputBufferLength;
                }
                status = STATUS_SUCCESS;
            }
            else
            {
                TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "NO IRQ; IOCTL_IRQ_WAIT_BEGIN Pending");
                deviceContext->COSSize = (ULONG)OutputBufferLength;
                deviceContext->PendingIrp = Request;
                deviceContext->HasPendingIrp = TRUE;
                WdfRequestForwardToIoQueue(Request, deviceContext->IRQQueue);
                deviceContext->InterruptRequested++;
                status = STATUS_PENDING;   // do not complete Irp of course
                //WdfRequestComplete(Request, status);
                return;
            }
        break;
    }						// IOCTL_IRQ_WAIT_BEGIN

    case IOCTL_IRQ_WAIT_ABORT:				// code == 0x812
        TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "IOCTL_IRQ_WAIT_ABORT");
        //{
        //    PDEVICE_CONTEXT deviceContext = AIOWDM4DeviceGetContext(WdfIoQueueGetDevice(Queue));
        //    if (!deviceContext->bRealDevice)
        //    {
        //        status = STATUS_OBJECT_TYPE_MISMATCH;
        //        break;
        //    }

        //    if (!deviceContext->InterruptRequested)
        //        status = STATUS_INVALID_DEVICE_REQUEST;
        //    else
        //    {
        //        deviceContext->InterruptOccurred = FALSE;
        //        deviceContext->InterruptRequested = FALSE;
        //        deviceContext->HasPendingIrp = FALSE;
        //        deviceContext->PendingIrp->IoStatus.Status = STATUS_CANCELLED;
        //        deviceContext->PendingIrp->IoStatus.Information = 0;
        //        IoSetCancelRoutine(deviceContext->PendingIrp, NULL);
        //        IoCompleteRequest(deviceContext->PendingIrp, IO_SERIAL_INCREMENT);
        status = STATUS_SUCCESS;
        //    }

        break;
        //} // case IOCTL_IRQ_WAIT_ABORT

    case IOCTL_FORGET_IRP:				//code == 0x880
        TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "IOCTL_FORGET_IRP");
        //{
        //    PDEVICE_CONTEXT deviceContext = AIOWDM4DeviceGetContext(WdfIoQueueGetDevice(Queue));
        //    if (!deviceContext->bRealDevice)
        //        status = STATUS_OBJECT_TYPE_MISMATCH;
        //    else if (!deviceContext->InterruptRequested)
        //        status = STATUS_NO_SUCH_FILE;
        //    else
        //    {
        //        deviceContext->InterruptRequested = FALSE;
        //        deviceContext->HasPendingIrp = FALSE;
        //        ForgetIRP(deviceContext->PendingIrp);
        //        deviceContext->PendingIrp = NULL;
        status = STATUS_SUCCESS;
        //    }

        break;
        //}

    case IOCTL_IRQ_POLL:				// code == 0x814
    {
        PDEVICE_CONTEXT deviceContext = AIOWDM4DeviceGetContext(WdfIoQueueGetDevice(Queue));
        if (!deviceContext->bRealDevice)
        {
            status = STATUS_OBJECT_TYPE_MISMATCH;
            break;
        }

        //if ( InterlockedExchange(PLONG(&(pdx->InterruptOccurred)), 0) )
        if (deviceContext->InterruptOccurred)
        {
            deviceContext->InterruptOccurred = 0; // --J2H Todo: turn into per-handle flag
            status = STATUS_SUCCESS;
        }
        else
            status = STATUS_NO_MORE_FILES
            ;

        break;
    }

    default:
    {
        TraceEvents(TRACE_LEVEL_INFORMATION,
            TRACE_QUEUE,
            "%!FUNC! Unknown IOCTL OutputBufferLength %d InputBufferLength %d IoControlCode %lx",
            (int)OutputBufferLength, (int)InputBufferLength, IoControlCode);

        status = STATUS_INVALID_DEVICE_REQUEST;
        break;
    }
    } // switch on IOCTL code

    WdfRequestCompleteWithInformation(Request, status, infolen);

    TraceEvents(TRACE_LEVEL_VERBOSE, TRACE_QUEUE, "<-- %!FUNC! Exit");
    return;
}

VOID
AIOWDM4EvtIoStop(
    _In_ WDFQUEUE Queue,
    _In_ WDFREQUEST Request,
    _In_ ULONG ActionFlags
)
/*++

Routine Description:

    This event is invoked for a power-managed queue before the device leaves the working state (D0).

Arguments:

    Queue -  Handle to the framework queue object that is associated with the
             I/O request.

    Request - Handle to a framework request object.

    ActionFlags - A bitwise OR of one or more WDF_REQUEST_STOP_ACTION_FLAGS-typed flags
                  that identify the reason that the callback function is being called
                  and whether the request is cancelable.

Return Value:

    VOID

--*/
{
    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "--> %!FUNC! Enter");

    TraceEvents(TRACE_LEVEL_VERBOSE,
        TRACE_QUEUE,
        "%!FUNC! Queue 0x%p, Request 0x%p ActionFlags %d",
        Queue, Request, ActionFlags);

    //
    // In most cases, the EvtIoStop callback function completes, cancels, or postpones
    // further processing of the I/O request.
    //
    // Typically, the driver uses the following rules:
    //
    // - If the driver owns the I/O request, it calls WdfRequestUnmarkCancelable
    //   (if the request is cancelable) and either calls WdfRequestStopAcknowledge
    //   with a Requeue value of TRUE, or it calls WdfRequestComplete with a
    //   completion status value of STATUS_SUCCESS or STATUS_CANCELLED.
    //
    //   Before it can call these methods safely, the driver must make sure that
    //   its implementation of EvtIoStop has exclusive access to the request.
    //
    //   In order to do that, the driver must synchronize access to the request
    //   to prevent other threads from manipulating the request concurrently.
    //   The synchronization method you choose will depend on your driver's design.
    //
    //   For example, if the request is held in a shared context, the EvtIoStop callback
    //   might acquire an internal driver lock, take the request from the shared context,
    //   and then release the lock. At this point, the EvtIoStop callback owns the request
    //   and can safely complete or requeue the request.
    //
    // - If the driver has forwarded the I/O request to an I/O target, it either calls
    //   WdfRequestCancelSentRequest to attempt to cancel the request, or it postpones
    //   further processing of the request and calls WdfRequestStopAcknowledge with
    //   a Requeue value of FALSE.
    //
    // A driver might choose to take no action in EvtIoStop for requests that are
    // guaranteed to complete in a small amount of time.
    //
    // In this case, the framework waits until the specified request is complete
    // before moving the device (or system) to a lower power state or removing the device.
    // Potentially, this inaction can prevent a system from entering its hibernation state
    // or another low system power state. In extreme cases, it can cause the system
    // to crash with bugcheck code 9F.
    //
    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_QUEUE, "<-- %!FUNC! Exit");
    return;
}
